@@ -36,18 +36,42 @@ setLocalTranslate('cl_lib3', sys.modules[__name__])
 class GitError(AddonError):
     """Git Error"""
 
+
 class NotGitError(GitError):
     """Данная директория не является Git"""
 
-class Git:
+
+class Git(object):
     """
     Объект для управление git репозиторием
     """
     skip_files = ["metadata/md5-cache",
                   "metadata/cache"]
 
-    def __init__(self):
+    def __init__(self, sshkey=None):
         self._git = self.__getGit()
+        self.sshkey = sshkey
+
+    def sshkey_env(self):
+        options = {
+            "UserKnownHostsFile": "/dev/null",
+            "StrictHostKeyChecking": "no",
+            "KbdInteractiveAuthentication": "no",
+            "ControlPath": "none",
+            "ControlMaster": "no",
+            "PasswordAuthentication": "no",
+            "BatchMode": "yes",
+            "PreferredAuthentications": "publickey"
+        }
+        if self.sshkey:
+            keyparam = "-i %s" % self.sshkey
+        else:
+            keyparam = ""
+        return {
+            "GIT_SSH_COMMAND": "ssh %s %s" % (
+                keyparam,
+                " ".join("-o %s=%s" % (k, v) for k, v in options.items()))
+        }
 
     def checkExistsRep(self, rpath):
         """
@@ -76,24 +100,51 @@ class Git:
 
     @staticmethod
     def _gitDir(rpath):
-        return path.join(rpath, ".git")
+        return path.abspath(path.join(rpath, ".git"))
 
     @staticmethod
     def is_git(gitpath):
         return path.isdir(Git._gitDir(gitpath))
 
-    def get_url(self, rpath, remote_name):
+    re_url = re.compile(r"^(ssh://(?:\w+@)?\w[\w\.-]+(?::\d+)?/[\w\./-]+?|"
+        r"https?://\w[\w\.-]+(?::\d+)?/[\w\./-]+?|"
+        r"((git://)?\w[\w\.-]+(?::\d+)?/[\w\./-]+?)|"
+        r"\w+@\w[\w\.-]+:[\w\./-]+?)(.git)?$")
+
+    @classmethod
+    def normalize_url(cls, url):
+        match = cls.re_url.match(url)
+        if not match:
+            return None
+        if match.group(2) and not match.group(3):
+            url = "git://%s" % match.group(2)
+        else:
+            url = match.group(1)
+        return "%s.git" % url
+
+    @staticmethod
+    def get_url(rpath, remote_name):
         config = ConfigParser(strict=False)
         config.read(path.join(Git._gitDir(rpath), "config"), encoding="utf-8")
         res = config.get('remote "%s"' % remote_name, "url",
                          raw=True, fallback="")
         if res:
             return res
-        fetch_info = self._getFetchHeadInfo(Git._gitDir(rpath))
+        fetch_info = Git.getFetchHeadInfo(Git._gitDir(rpath))
         if "url" in fetch_info:
             return fetch_info['url']
         return ""
 
+    def process(self, *args, **kw):
+        kw["envdict"] = os.environ.copy()
+        kw["envdict"].update(self.sshkey_env())
+        return process(*args, **kw)
+
+    def process_progress(self, *args, **kw):
+        kw["envdict"] = os.environ.copy()
+        kw["envdict"].update(self.sshkey_env())
+        return PercentProgress(*args, **kw)
+
     def cloneRepository(self, url, rpath, branch, cb_progress=None):
         """
         Сделать локальную копию репозитория
@@ -104,23 +155,26 @@ class Git:
             branch: ветка на которую необходимо переключиться
         """
         if cb_progress:
-            gitClone = PercentProgress(self._git, "clone", "-q",
-                                       "--no-single-branch", "--progress",
-                                       "--verbose",
-                                       "--depth=1", "-b", branch, url, rpath,
-                                       part=4, stderr=STDOUT)
+            gitClone = self.process_progress(self._git, "clone", "-q",
+                                             "--no-single-branch", "--progress",
+                                             "--verbose",
+                                             "--depth=1", "-b", branch, url,
+                                             rpath, part=4, stderr=STDOUT)
             for perc in gitClone.progress():
                 cb_progress(perc)
         else:
-            gitClone = process(self._git, "clone", "-q", "--no-single-branch",
-                               "--depth=1", "-b", branch, url, rpath,
-                               stderr=STDOUT)
+            gitClone = self.process(
+                self._git, "clone", "-q", "--no-single-branch",
+                "--depth=1", "-b", branch, url, rpath,
+                stderr=STDOUT)
         if gitClone.failed():
             error = gitClone.read()
             if "Remote branch %s not found" % branch in error:
                 raise GitError(
                     _("Branch {branch} not found in repository {url}").format(
                         branch=branch, url=url))
+            if error:
+                self.detect_error(error.split("\n"), url)
             raise GitError(_("Failed to clone repository {url}").format(
                 url=url), error)
         return True
@@ -133,13 +187,13 @@ class Git:
                 wholeCommand = [self._git, "--git-dir", git_dir,
                                 "--work-tree", rpath] + list(args)
             if cb_progress and args[0] in ("reset", "fetch", "checkout"):
-                gitCmd = PercentProgress(
+                gitCmd = self.process_progress(
                     *wholeCommand,
                     stderr=STDOUT, atty=True, **kw)
                 for perc in gitCmd.progress():
                     cb_progress(perc)
             else:
-                gitCmd = process(*wholeCommand, stderr=STDOUT)
+                gitCmd = self.process(*wholeCommand, stderr=STDOUT)
             if gitCmd.failed():
                 error.append(gitCmd.read())
                 return False
@@ -147,7 +201,8 @@ class Git:
 
         return command
 
-    def _getFetchHeadInfo(self, git_dir):
+    @staticmethod
+    def getFetchHeadInfo(git_dir):
         data = readFile(path.join(git_dir, "FETCH_HEAD")).strip()
         re_fetch = re.compile(r"^(\w+)\s+(?:(branch|tag) '([^']+)' of )?(\S+)$")
         m = re_fetch.search(data)
@@ -155,9 +210,9 @@ class Git:
             d = {'commit': m.group(1),
                  'url': m.group(4)}
             if m.group(2):
-                type_map = {'tag': self.Reference.Tag,
-                            'branch': self.Reference.Branch}
-                d['type'] = type_map.get(m.group(2), self.Reference.Unknown)
+                type_map = {'tag': Git.Reference.Tag,
+                            'branch': Git.Reference.Branch}
+                d['type'] = type_map.get(m.group(2), Git.Reference.Unknown)
                 d['reference'] = m.group(3)
             return d
         return {}
@@ -181,6 +236,23 @@ class Git:
                 pass
         return False
 
+    def detect_error(self, error, url):
+        error_map = {
+            "access denied": _("Access denied to repository {url}"),
+            "Permission denied": _("Access denied to repository {url}"),
+            "does not appear to be a git repository":
+                _("{url} does not appear to be a git "
+                  "repository"),
+            "Could not read from remote":
+                _("Could not read from remote repository {url}")
+        }
+        if error:
+            if self.sshkey and "ENCRYPTED" in readFile(self.sshkey):
+                raise GitError("Private key is locked")
+            for token, errmess in error_map.items():
+                if token in error[-1]:
+                    raise GitError(errmess.format(url=url))
+
     def cloneTagRepository(self, url, rpath, reference, cb_progress=None,
                            copy_index=False):
         """
@@ -210,6 +282,8 @@ class Git:
                 if self._clear_lock(git_dir, "shallow.lock", error):
                     res = command("fetch", "--depth=1", url, reference, **kw)
             if not res:
+                if error:
+                    self.detect_error(error, url)
                 raise GitError(
                     _("Reference {mark} not found in repository {url}"
                       ).format(mark=reference, url=url))
@@ -239,6 +313,8 @@ class Git:
                         command("symbolic-ref", "HEAD",
                                 "refs/heads/%s" % reference)
                     return True
+        if error:
+            self.detect_error(error, url)
         raise GitError(_("Failed to clone repository {url}").format(
             url=url), error[-1])
         # progressParam = {'fetch': {'part': 4, 'end': False},
@@ -249,7 +325,7 @@ class Git:
         Удалить в репозитории лишнии объекты
         """
         git_dir = self._gitDir(rpath)
-        head_info = self._getFetchHeadInfo(git_dir)
+        head_info = self.getFetchHeadInfo(git_dir)
         if not head_info or not head_info.get('reference', ''):
             return False
 
@@ -277,7 +353,7 @@ class Git:
         :return: True репозиторий переключается по тэгам
         """
         git_path = self._gitDir(rpath)
-        data = self._getFetchHeadInfo(git_path)
+        data = self.getFetchHeadInfo(git_path)
         if "type" in data and data['type'] == self.Reference.Tag:
             return True
         return False
@@ -353,9 +429,11 @@ class Git:
                 res = command("fetch", "--depth=1",
                               url, reference, part=4, end=False)
         if not res:
+            if error:
+                self.detect_error(error, url)
             raise GitError(
-                _("Reference {mark} not found in repository {url}"
-                  ).format(mark=reference, url=url))
+                _("Reference {mark} not found in repository {url}").format(
+                    mark=reference, url=url))
         res = command("reset", "--hard", "FETCH_HEAD", part=4, startpart=3)
         if not res:
             if self._clear_lock(git_dir, "index.lock", error):
@@ -367,6 +445,8 @@ class Git:
             if self.saveFetchHead(git_dir) == self.Reference.Branch:
                 command("symbolic-ref", "HEAD", "refs/heads/%s" % reference)
             return True
+        if error:
+            self.detect_error(error, url)
         raise GitError(_("Failed to clone repository {url}").format(
             url=url), error[-1])
 
@@ -378,7 +458,7 @@ class Git:
         :param git_dir: каталог .git репозитория
         :return: Reference
         """
-        fetch_info = self._getFetchHeadInfo(git_dir)
+        fetch_info = self.getFetchHeadInfo(git_dir)
         if force_type is not None:
             fetch_info['type'] = force_type
         if force_ref is not None:
@@ -406,7 +486,8 @@ class Git:
         Rsync = 873
         Unknown = 0
 
-    def _parse_url(self, url):
+    @staticmethod
+    def parse_url(url):
         """
         Разобрать url на составляющие: тип и хост
 
@@ -414,27 +495,29 @@ class Git:
         :return: (тип, хост)
         """
         net_protocols = {
-            'ssh': self.GitProtocol.SSH,
-            'ssh+git': self.GitProtocol.SSH,
-            'git+ssh': self.GitProtocol.SSH,
-            'http': self.GitProtocol.HTTP,
-            'https': self.GitProtocol.HTTPS,
-            'ftp': self.GitProtocol.FTP,
-            'ftps': self.GitProtocol.FTPS,
-            'rsync': self.GitProtocol.Rsync,
-            'git': self.GitProtocol.Git}
+            'ssh': Git.GitProtocol.SSH,
+            'ssh+git': Git.GitProtocol.SSH,
+            'git+ssh': Git.GitProtocol.SSH,
+            'http': Git.GitProtocol.HTTP,
+            'https': Git.GitProtocol.HTTPS,
+            'ftp': Git.GitProtocol.FTP,
+            'ftps': Git.GitProtocol.FTPS,
+            'rsync': Git.GitProtocol.Rsync,
+            'git': Git.GitProtocol.Git}
         re_host = re.compile(
-            "(%s)://(?:[^@]+@)?([^:/]+)" %
+            "(%s)://(?:[^@]+@)?([^:/]+)(:\d+)?" %
             "|".join(x.replace("+", r"\+") for x in net_protocols.keys())
         )
         net_match = re_host.search(url)
         if net_match:
+            if net_match.group(3):
+                return int(net_match.group(3)), net_match.group(2)
             return net_protocols[net_match.group(1)], net_match.group(2)
         re_scp = re.compile("(?:[^@]+@)?([^:]+):(//)?")
         scp_match = re_scp.search(url)
         if scp_match and scp_match.group(2) is None:
-            return self.GitProtocol.SSH, scp_match.group(1)
-        return self.GitProtocol.Unknown, ""
+            return Git.GitProtocol.SSH, scp_match.group(1)
+        return Git.GitProtocol.Unknown, ""
 
     def checkUrl(self, url):
         """
@@ -453,7 +536,7 @@ class Git:
         :param url: проверяемый URL
         :return:
         """
-        git_port, target = self._parse_url(url)
+        git_port, target = Git.parse_url(url)
         if git_port != self.GitProtocol.Unknown:
             return check_port(target, git_port, timeout=20)
         return True
@@ -471,9 +554,9 @@ class Git:
                 return False
         else:
             git_dir = self._gitDir(rpath)
-            git_log = process(self._git, "--git-dir", git_dir, "log",
-                              "-n1", "--format=format:%H",
-                              "--quiet", "--", *fns, stderr=STDOUT)
+            git_log = self.process(self._git, "--git-dir", git_dir, "log",
+                                   "-n1", "--format=format:%H",
+                                   "--quiet", "--", *fns, stderr=STDOUT)
             if git_log.success() and git_log.read().strip():
                 return True
             status = self.getStatusInfo(rpath)
@@ -539,13 +622,13 @@ class Git:
             if cb_progress and commands[cmd][0] in ("fetch", "checkout"):
                 progressParam = {'fetch': {'part': 4, 'end': False},
                                  'checkout': {'part': 4, 'startpart': 3}}
-                gitCmd = PercentProgress(
+                gitCmd = self.process_progress(
                     *wholeCommand + ["--progress", "--verbose"],
                     stderr=STDOUT, **progressParam)
                 for perc in gitCmd.progress():
                     cb_progress(perc)
             else:
-                gitCmd = process(*wholeCommand, stderr=STDOUT)
+                gitCmd = self.process(*wholeCommand, stderr=STDOUT)
             if gitCmd.failed():
                 error.append(gitCmd.read())
                 return False
@@ -565,6 +648,8 @@ class Git:
                         return True
                 elif command("checkout"):
                     return False
+        if error:
+            self.detect_error(error, url)
         raise GitError(_("Failed to clone repository {url}").format(
             url=url), error[-1])
 
@@ -572,8 +657,8 @@ class Git:
         """
         Обновить репозиторий до последней версии
         """
-        gitPull = process(self._git, "--git-dir", self._gitDir(rpath),
-                          "pull", "--ff-only", stderr=STDOUT)
+        gitPull = self.process(self._git, "--git-dir", self._gitDir(rpath),
+                               "pull", "--ff-only", stderr=STDOUT)
         if gitPull.failed():
             if not quiet_error:
                 error = gitPull.read()
@@ -588,15 +673,15 @@ class Git:
         Получить изменения из удаленно репозитория
         """
         if cb_progress:
-            gitFetch = PercentProgress(self._git, "--git-dir",
-                                       self._gitDir(rpath),
-                                       "fetch", "--progress", "--verbose",
-                                       part=3, stderr=STDOUT)
+            gitFetch = self.process_progress(self._git, "--git-dir",
+                                             self._gitDir(rpath),
+                                             "fetch", "--progress", "--verbose",
+                                             part=3, stderr=STDOUT)
             for perc in gitFetch.progress():
                 cb_progress(perc)
         else:
-            gitFetch = process(self._git, "--git-dir", self._gitDir(rpath),
-                               "fetch", stderr=STDOUT)
+            gitFetch = self.process(self._git, "--git-dir", self._gitDir(rpath),
+                                    "fetch", stderr=STDOUT)
         if gitFetch.failed():
             error = gitFetch.read()
             raise GitError(
@@ -609,9 +694,10 @@ class Git:
         Проверить наличие изменений пользователем файлов в репозитории
         """
         git_dir = self._gitDir(rpath)
-        git_status = process(self._git, "--git-dir", git_dir, "--work-tree",
-                             rpath,
-                             "status", "--porcelain", stderr=STDOUT)
+        git_status = self.process(self._git, "--git-dir", git_dir,
+                                  "--work-tree",
+                                  rpath,
+                                  "status", "--porcelain", stderr=STDOUT)
         if git_status.success():
             return not any(x.strip() for x in git_status)
         else:
@@ -657,9 +743,9 @@ class Git:
 
     def getCommit(self, rpath, reference):
         git_dir = self._gitDir(rpath)
-        git_show = process(self._git, "--git-dir", git_dir, "log",
-                           "-n1", "--format=format:%H",
-                           "--quiet", reference, stderr=STDOUT)
+        git_show = self.process(self._git, "--git-dir", git_dir, "log",
+                                "-n1", "--format=format:%H",
+                                "--quiet", reference, stderr=STDOUT)
         if git_show.success():
             return git_show.read().strip()
         else:
@@ -675,9 +761,10 @@ class Git:
             Словарь выдаваемый функцией _parseStatusInfo
         """
         git_dir = self._gitDir(rpath)
-        git_status = process(self._git, "--git-dir", git_dir, "--work-tree",
-                             rpath,
-                             "status", "-b", "--porcelain", stderr=STDOUT)
+        git_status = self.process(self._git, "--git-dir", git_dir,
+                                  "--work-tree",
+                                  rpath,
+                                  "status", "-b", "--porcelain", stderr=STDOUT)
         if git_status.success():
             retDict = self.parseStatusInfo(git_status.read())
             if not retDict:
@@ -714,12 +801,12 @@ class Git:
                              "origin/%s" % info["branch"] == info["origin"])):
             return False
         commit = (info['origin'] if to_origin else to_rev) or "HEAD"
-        git_reset = process(self._git, "--git-dir", git_dir, "--work-tree",
-                            rpath,
-                            "reset", "--hard", commit, stderr=STDOUT)
-        git_clean = process(self._git, "--git-dir", git_dir, "--work-tree",
-                            rpath,
-                            "clean", "-fd", stderr=STDOUT)
+        git_reset = self.process(self._git, "--git-dir", git_dir, "--work-tree",
+                                 rpath,
+                                 "reset", "--hard", commit, stderr=STDOUT)
+        git_clean = self.process(self._git, "--git-dir", git_dir, "--work-tree",
+                                 rpath,
+                                 "clean", "-fd", stderr=STDOUT)
         if git_reset.failed() or git_clean.failed():
             raise GitError(_("Failed to clean the {rpath} repository").format(
                 rpath=rpath))
@@ -736,9 +823,9 @@ class Git:
         Получить список тэгов, отсортированных по версии
         """
         git_dir = self._gitDir(rpath)
-        git_tag = process(self._git, "--git-dir", git_dir, "--work-tree",
-                          rpath,
-                          "tag", "--sort=version:refname", stderr=STDOUT)
+        git_tag = self.process(self._git, "--git-dir", git_dir, "--work-tree",
+                               rpath,
+                               "tag", "--sort=version:refname", stderr=STDOUT)
         if git_tag.success():
             return [x.strip() for x in git_tag if x.strip()]
         return []
@@ -748,9 +835,9 @@ class Git:
         Сменить ветку
         """
         git_dir = self._gitDir(rpath)
-        git_checkout = process(self._git, "--git-dir", git_dir,
-                               "--work-tree", rpath,
-                               "checkout", "-f", branch, stderr=STDOUT)
+        git_checkout = self.process(self._git, "--git-dir", git_dir,
+                                    "--work-tree", rpath,
+                                    "checkout", "-f", branch, stderr=STDOUT)
         if git_checkout.failed():
             error = git_checkout.read()
             if "pathspec '%s' did not match" % branch in error:
